import struct

SL_GATEWAY_IP = "ip"
SL_GATEWAY_PORT = "port"
SL_GATEWAY_TYPE = "gtype"
SL_GATEWAY_SUBTYPE = "gsubtype"
SL_GATEWAY_NAME = "name"


class ScreenLogicError(Exception):
    pass


class ScreenLogicWarning(Exception):
    pass


class header:
    fmt = "<HHI"
    length = struct.calcsize(fmt)


# Some of the message codes
class code:
    MSG_CODE_1 = 0
    UNKNOWN_ANSWER = 13
    CHALLENGE_QUERY = 14
    CHALLENGE_ANSWER = CHALLENGE_QUERY + 1
    LOCALLOGIN_QUERY = 27
    LOCALLOGIN_ANSWER = LOCALLOGIN_QUERY + 1
    VERSION_QUERY = 8120
    VERSION_ANSWER = VERSION_QUERY + 1
    POOLSTATUS_QUERY = 12526
    POOLSTATUS_ANSWER = POOLSTATUS_QUERY + 1
    SETHEATTEMP_QUERY = 12528
    SETHEATTEMP_ANSWER = SETHEATTEMP_QUERY + 1
    BUTTONPRESS_QUERY = 12530
    BUTTONPRESS_ANSWER = BUTTONPRESS_QUERY + 1
    CTRLCONFIG_QUERY = 12532
    CTRLCONFIG_ANSWER = CTRLCONFIG_QUERY + 1
    SETHEATMODE_QUERY = 12538
    SETHEATMODE_ANSWER = SETHEATMODE_QUERY + 1
    SETCOOLTEMP_QUERY = 12590
    SETCOOLTEMP_ANSWER = SETCOOLTEMP_QUERY + 1
    GATEWAYDATA_QUERY = 18003
    GATEWAYDATA_ANSWER = GATEWAYDATA_QUERY + 1
    CONTROLLER_QUERY = 12532
    CONTROLLER_ANSWER = CONTROLLER_QUERY + 1
    EQUIPMENT_QUERY = 12566
    EQUIPMENT_ANSWER = EQUIPMENT_QUERY + 1
    PUMPSTATUS_QUERY = 12584
    PUMPSTATUS_ANSWER = PUMPSTATUS_QUERY + 1
    LIGHTCOMMAND_QUERY = 12556
    LIGHTCOMMAND_ANSWER = LIGHTCOMMAND_QUERY + 1
    CHEMISTRY_QUERY = 12592
    CHEMISTRY_ANSWER = CHEMISTRY_QUERY + 1
    SCGCONFIG_QUERY = 12572
    SCGCONFIG_ANSWER = SCGCONFIG_QUERY + 1


class DATA:
    KEY_ALERTS = "alerts"
    KEY_BODIES = "bodies"
    KEY_CHEMISTRY = "chemistry"
    KEY_CIRCUITS = "circuits"
    KEY_COLORS = "colors"
    KEY_CONFIG = "config"
    KEY_NOTIFICATIONS = "notifications"
    KEY_PUMPS = "pumps"
    KEY_SCG = "scg"
    KEY_SENSORS = "sensors"


# class mapping:
#    BODY_TYPE  = ['Pool', 'Spa']

#    HEAT_MODE  = ['Off', 'Solar',
#                  "Solar Preferred",
#                  'Heat', "Don't Change"]

#    ON_OFF     = ['Off', 'On']

#    COLOR_MODE = ['Off', 'On',
#                  'Set', 'Sync',
#                  'Swim', 'Party',
#                  'Romantic', 'Caribbean',
#                  'American', 'Sunset',
#                  'Royal', 'Save',
#                  'Recall', 'Blue',
#                  'Green', 'Red',
#                  'White', 'Magenta',
#                  'Thumper', 'Next',
#                  'Reset', 'Hold']


class BODY_TYPE:
    POOL = 0
    SPA = 1

    NAME_FOR_NUM = {POOL: "Pool", SPA: "Spa"}

    NUM_FOR_NAME = {name: num for num, name in NAME_FOR_NUM.items()}


class ON_OFF:
    OFF = 0
    ON = 1

    NAME_FOR_NUM = {OFF: "Off", ON: "On"}

    NUM_FOR_NAME = {name: num for num, name in NAME_FOR_NUM.items()}

    @classmethod
    def from_bool(cls, expresson: bool):
        return cls.ON if expresson else cls.OFF


class HEAT_MODE:
    OFF = 0
    SOLAR = 1
    SOLAR_PREFERRED = 2
    HEATER = 3
    DONT_CHANGE = 4

    NAME_FOR_NUM = {
        OFF: "Off",
        SOLAR: "Solar",
        SOLAR_PREFERRED: "Solar Preferred",
        HEATER: "Heater",
        DONT_CHANGE: "Don't Change",
    }

    NUM_FOR_NAME = {name: num for num, name in NAME_FOR_NUM.items()}


class COLOR_MODE:
    OFF = 0
    ON = 1
    SET = 2
    SYNC = 3
    SWIM = 4
    PARTY = 5
    ROMANCE = 6
    CARIBBEAN = 7
    AMERICAN = 8
    SUNSET = 9
    ROYAL = 10
    SAVE = 11
    RECALL = 12
    BLUE = 13
    GREEN = 14
    RED = 15
    WHITE = 16
    MAGENTA = 17
    THUMPER = 18
    NEXT = 19
    RESET = 20
    HOLD = 21

    NAME_FOR_NUM = {
        OFF: "All Off",
        ON: "All On",
        SET: "Color Set",
        SYNC: "Color Sync",
        SWIM: "Color Swim",
        PARTY: "Party",
        ROMANCE: "Romance",
        CARIBBEAN: "Caribbean",
        AMERICAN: "American",
        SUNSET: "Sunset",
        ROYAL: "Royal",
        SAVE: "Save",
        RECALL: "Recall",
        BLUE: "Blue",
        GREEN: "Green",
        RED: "Red",
        WHITE: "White",
        MAGENTA: "Magenta",
        THUMPER: "Thumper",
        NEXT: "Next Mode",
        RESET: "Reset",
        HOLD: "Hold",
    }

    NUM_FOR_NAME = {name: num for num, name in NAME_FOR_NUM.items()}


class DEVICE_TYPE:
    ALARM = "alarm"
    ENERGY = "energy"
    TEMPERATURE = "temperature"


class UNIT:
    CELSIUS = "\xb0C"
    FAHRENHEIT = "\xb0F"


class EQUIPMENT:
    CONTROLLER_HARDWARE = {
        0: {0: "IntelliTouch i5+3S"},
        1: {0: "IntelliTouch i7+3"},
        2: {0: "IntelliTouch i9+3"},
        3: {0: "IntelliTouch i5+3S"},
        4: {0: "IntelliTouch i9+3S"},
        5: {0: "IntelliTouch i10+3D"},
        11: {0: "Suntouch/Intellicom"},
        13: {
            0: "EasyTouch2 8",
            1: "EasyTouch2 8P",
            2: "EasyTouch2 4",
            3: "EasyTouch2 4P",
            5: "EasyTouch2 PL4",
            6: "EasyTouch2 PSL4",
        },
        14: {
            0: "EasyTouch1 8",
            1: "EasyTouch1 8P",
            2: "EasyTouch1 4",
            3: "EasyTouch1 4P",
        },
    }

    PUMP_TYPE = {
        0: "None",
        1: "Intelliflow VF",
        2: "Intelliflow VS",
        3: "Intelliflow VSF",
    }

    FLAG_SOLAR = 0x1
    FLAG_SOLAR_AS_HEAT_PUMP = 0x2
    FLAG_CHLORINATOR = 0x4
    FLAG_SPA_SIDE_REMOTE = 0x20
    FLAG_COOLING = 0x800
    FLAG_INTELLICHEM = 0x8000


class CIRCUIT_FUNCTION:
    # Known circuit functions.
    GENERIC = 0
    SPA = 1
    POOL = 2
    MASTER_CLEANER = 5
    LIGHT = 7
    MAGICSTREAM = 8  # ?
    SPILLWAY = 14
    INTELLIBRITE = 16


class INTERFACE_GROUP:
    # Known interface groups
    POOL = 0
    SPA = 1
    FEATURES = 2
    LIGHTS = 3


class CHEMISTRY:
    FLAG_ALARM_FLOW = 0x01
    FLAG_ALARM_PH = 0x06
    FLAG_ALARM_ORP = 0x18
    FLAG_ALARM_PH_SUPPLY = 0x20
    FLAG_ALARM_ORP_SUPPLY = 0x40
    FLAG_ALARM_PROBE_FAULT = 0x80

    FLAG_STATUS_NORMAL = 0x01
    FLAG_STATUS_IDEAL = 0x02
    FLAG_STATUS_CORROSIVE = 0x04
    FLAG_STATUS_SCALING = 0x08

    MASK_STATUS_PH_DOSING = 0x30
    MASK_STATUS_ORP_DOSING = 0xC0

    FLAG_WARNING_PH_LOCKOUT = 0x01
    FLAG_WARNING_PH_LIMIT = 0x02
    FLAG_WARNING_ORP_LIMIT = 0x04


class CHEM_DOSING_STATE:
    DOSING = 0
    MIXING = 1
    MONITORING = 2

    NAME_FOR_NUM = {DOSING: "Dosing", MIXING: "Mixing", MONITORING: "Monitoring"}
    NUM_FOR_NAME = {name: num for num, name in NAME_FOR_NUM.items()}


GENERIC_CIRCUIT_NAMES = [
    *[f"Aux {num}" for num in range(1, 8)],  # Last number is 7
    "AuxEx",
    *[f"Feature {num}" for num in range(1, 9)],  # Last number is 8
]

DEFAULT_CIRCUIT_NAMES = ["Spa", "Pool", *GENERIC_CIRCUIT_NAMES]


# COLOR_MODES_* may not be complete
COLOR_MODES_GENERIC = {num: COLOR_MODE.NAME_FOR_NUM[num] for num in [0, 1]}

COLOR_MODES_COLORS = {num: COLOR_MODE.NAME_FOR_NUM[num] for num in [13, 14, 15, 16, 17]}

COLOR_MODES_SAM = {
    **COLOR_MODES_GENERIC,
    **{num: COLOR_MODE.NAME_FOR_NUM[num] for num in [2, 3, 4]},
    **COLOR_MODES_COLORS,
}

COLOR_MODES_INTELLIBRITE = {
    **COLOR_MODES_GENERIC,
    **COLOR_MODES_SAM,
    **{num: COLOR_MODE.NAME_FOR_NUM[num] for num in [5, 6, 7, 8, 9, 10, 11, 12]},
    **COLOR_MODES_COLORS,
}

COLOR_MODES_MAGICSTREAM = {
    **COLOR_MODES_GENERIC,
    **{num: COLOR_MODE.NAME_FOR_NUM[num] for num in [18, 19, 20, 21]},
}



import struct
from .utility import sendReceiveMessage
from ..const import code


def request_pool_button_press(gateway_socket, circuit_id, circuit_state):
    response = sendReceiveMessage(
        gateway_socket,
        code.BUTTONPRESS_QUERY,
        struct.pack("<III", 0, circuit_id, circuit_state),
    )
    if response == b"":
        return True
    else:
        return False



from .utility import sendReceiveMessage, decodeMessageString
from ..const import code


def request_gateway_version(gateway_socket):
    response = sendReceiveMessage(gateway_socket, code.VERSION_QUERY)
    return decodeMessageString(response)



import time
import struct
import socket
from .utility import sendReceiveMessage, encodeMessageString, decodeMessageString
from ..const import code, ScreenLogicError


def create_login_message():
    # these constants are only for this message.
    schema = 348
    connectionType = 0
    clientVersion = encodeMessageString("Android")
    pid = 2
    password = "0000000000000000"  # passwd must be <= 16 chars. empty is not OK.
    passwd = encodeMessageString(password)
    fmt = "<II" + str(len(clientVersion)) + "s" + str(len(passwd)) + "sxI"
    return struct.pack(fmt, schema, connectionType, clientVersion, passwd, pid)


def create_socket(ip, port):
    tcpSock = None
    # pylint: disable=unused-variable
    for result in socket.getaddrinfo(ip, port, socket.AF_UNSPEC, socket.SOCK_STREAM):
        af, socktype, proto, canonname, sa = result
        try:
            tcpSock = socket.socket(af, socktype, proto)
        except OSError:
            tcpSock = None
            continue
        try:
            tcpSock.settimeout(5)
            tcpSock.connect(sa)
        except OSError:
            tcpSock.close()
            tcpSock = None
            continue
        break

    if not tcpSock:
        raise ScreenLogicError(f"Unable to connect to {ip}:{port}")

    return tcpSock


def gateway_connect(connected_socket):
    connectString = b"CONNECTSERVERHOST\r\n\r\n"  # as bytes, not string
    connected_socket.sendall(connectString)
    time.sleep(0.25)
    response = sendReceiveMessage(connected_socket, code.CHALLENGE_QUERY)
    mac_address = decodeMessageString(response)
    return mac_address


def gateway_login(connected_socket):
    msg = create_login_message()
    try:
        # Gateway will respond with the response code and that's all.
        _ = sendReceiveMessage(connected_socket, code.LOCALLOGIN_QUERY, msg)
        return True
    except ScreenLogicError:
        return False


def connect_to_gateway(gateway_ip, gateway_port):
    connected_socket = create_socket(gateway_ip, gateway_port)

    mac_address = gateway_connect(connected_socket)

    if gateway_login(connected_socket):
        return connected_socket, mac_address
    else:
        return None



# import json
import struct
from .utility import sendReceiveMessage, getSome
from ..const import code, DATA


def request_scg_config(gateway_socket, data):
    response = sendReceiveMessage(
        gateway_socket, code.SCGCONFIG_QUERY, struct.pack("<I", 0)
    )
    decode_scg_config(response, data)


def decode_scg_config(buff, data):
    # print(buff)

    if DATA.KEY_SCG not in data:
        data[DATA.KEY_SCG] = {}

    scg = data[DATA.KEY_SCG]

    present, offset = getSome("I", buff, 0)
    scg["scg_present"] = present

    status, offset = getSome("I", buff, offset)
    scg["scg_status"] = {"name": "SCG Status", "value": status}

    level1, offset = getSome("I", buff, offset)
    scg["scg_level1"] = {"name": "SCG Level 1", "value": level1, "unit": "%"}

    level2, offset = getSome("I", buff, offset)
    scg["scg_level2"] = {"name": "SCG Level 2", "value": level2, "unit": "%"}

    salt, offset = getSome("I", buff, offset)
    scg["scg_salt_ppm"] = {"name": "SCG Salt", "value": (salt * 50), "unit": "ppm"}

    flags, offset = getSome("I", buff, offset)
    scg["scg_flags"] = flags

    superChlorTimer, offset = getSome("I", buff, offset)
    scg["scg_super_chlor_timer"] = {
        "name": "SCG Super Chlorination Timer",
        "value": superChlorTimer,
    }

    # print(json.dumps(data, indent=2))



# import json
import struct
from .utility import sendReceiveMessage, getSome
from ..const import code, BODY_TYPE, DATA, DEVICE_TYPE, UNIT


def request_pool_status(gateway_socket, data):
    response = sendReceiveMessage(
        gateway_socket, code.POOLSTATUS_QUERY, struct.pack("<I", 0)
    )
    decode_pool_status(response, data)


# pylint: disable=unused-variable
def decode_pool_status(buff, data):
    # print(buff)

    if DATA.KEY_CONFIG not in data:
        data[DATA.KEY_CONFIG] = {}

    config = data[DATA.KEY_CONFIG]

    ok, offset = getSome("I", buff, 0)
    config["ok"] = {"name": "OK Check", "value": ok}

    freezeMode, offset = getSome("B", buff, offset)
    config["freeze_mode"] = {"name": "Freeze Mode", "value": freezeMode}

    remotes, offset = getSome("B", buff, offset)
    config["remotes"] = {"name": "Remotes", "value": remotes}

    poolDelay, offset = getSome("B", buff, offset)
    config["pool_delay"] = {"name": "Pool Delay", "value": poolDelay}

    spaDelay, offset = getSome("B", buff, offset)
    config["spa_delay"] = {"name": "Spa Delay", "value": spaDelay}

    cleanerDelay, offset = getSome("B", buff, offset)
    config["cleaner_delay"] = {"name": "Cleaner Delay", "value": cleanerDelay}

    # fast forward 3 bytes. Unknown data.
    ff1, offset = getSome("B", buff, offset)
    ff2, offset = getSome("B", buff, offset)
    ff3, offset = getSome("B", buff, offset)

    unit_txt = (
        UNIT.CELSIUS
        if "is_celsius" in config and config["is_celsius"]["value"]
        else UNIT.FAHRENHEIT
    )

    if DATA.KEY_SENSORS not in data:
        data[DATA.KEY_SENSORS] = {}

    sensors = data[DATA.KEY_SENSORS]

    airTemp, offset = getSome("i", buff, offset)
    sensors["air_temperature"] = {
        "name": "Air Temperature",
        "value": airTemp,
        "unit": unit_txt,
        "device_type": DEVICE_TYPE.TEMPERATURE,
    }

    bodiesCount, offset = getSome("I", buff, offset)
    # Should this default to 2?
    bodiesCount = min(bodiesCount, 2)

    if DATA.KEY_BODIES not in data:
        data[DATA.KEY_BODIES] = {}

    bodies = data[DATA.KEY_BODIES]

    for i in range(bodiesCount):
        bodyType, offset = getSome("I", buff, offset)
        if bodyType not in range(2):
            bodyType = 0

        if i not in bodies:
            bodies[i] = {}

        currentBody = bodies[i]

        if "min_set_point" not in currentBody:
            currentBody["min_set_point"] = {}

        currentBody["min_set_point"]["unit"] = unit_txt

        if "max_set_point" not in currentBody:
            currentBody["max_set_point"] = {}

        currentBody["max_set_point"]["unit"] = unit_txt

        currentBody["body_type"] = {"name": "Type of body of water", "value": bodyType}

        lastTemp, offset = getSome("i", buff, offset)
        bodyName = "Last {} Temperature".format(BODY_TYPE.NAME_FOR_NUM[bodyType])
        currentBody["last_temperature"] = {
            "name": bodyName,
            "value": lastTemp,
            "unit": unit_txt,
            "device_type": DEVICE_TYPE.TEMPERATURE,
        }

        heatStatus, offset = getSome("i", buff, offset)
        heaterName = "{} Heat".format(BODY_TYPE.NAME_FOR_NUM[bodyType])
        currentBody["heat_status"] = {"name": heaterName, "value": heatStatus}

        heatSetPoint, offset = getSome("i", buff, offset)
        hspName = "{} Heat Set Point".format(BODY_TYPE.NAME_FOR_NUM[bodyType])
        currentBody["heat_set_point"] = {
            "name": hspName,
            "value": heatSetPoint,
            "unit": unit_txt,
            "device_type": DEVICE_TYPE.TEMPERATURE,
        }

        coolSetPoint, offset = getSome("i", buff, offset)
        cspName = "{} Cool Set Point".format(BODY_TYPE.NAME_FOR_NUM[bodyType])
        currentBody["cool_set_point"] = {
            "name": cspName,
            "value": coolSetPoint,
            "unit": unit_txt,
        }

        heatMode, offset = getSome("i", buff, offset)
        hmName = "{} Heat Mode".format(BODY_TYPE.NAME_FOR_NUM[bodyType])
        currentBody["heat_mode"] = {"name": hmName, "value": heatMode}

    circuitCount, offset = getSome("I", buff, offset)

    if DATA.KEY_CIRCUITS not in data:
        data[DATA.KEY_CIRCUITS] = {}

    circuits = data[DATA.KEY_CIRCUITS]

    for i in range(circuitCount):
        circuitID, offset = getSome("I", buff, offset)

        if circuitID not in circuits:
            circuits[circuitID] = {}

        currentCircuit = circuits[circuitID]

        if "id" not in currentCircuit:
            currentCircuit["id"] = circuitID

        circuitstate, offset = getSome("I", buff, offset)
        currentCircuit["value"] = circuitstate

        cColorSet, offset = getSome("B", buff, offset)
        currentCircuit["color_set"] = cColorSet

        cColorPos, offset = getSome("B", buff, offset)
        currentCircuit["color_position"] = cColorPos

        cColorStagger, offset = getSome("B", buff, offset)
        currentCircuit["color_stagger"] = cColorStagger

        circuitDelay, offset = getSome("B", buff, offset)
        currentCircuit["delay"] = circuitDelay

    pH, offset = getSome("i", buff, offset)
    sensors["ph"] = {"name": "pH", "value": (pH / 100), "unit": "pH"}

    orp, offset = getSome("i", buff, offset)
    sensors["orp"] = {"name": "ORP", "value": orp, "unit": "mV"}

    saturation, offset = getSome("i", buff, offset)
    sensors["saturation"] = {
        "name": "Saturation Index",
        "value": (saturation / 100),
        "unit": "lsi",
    }

    saltPPM, offset = getSome("i", buff, offset)
    sensors["salt_ppm"] = {"name": "Salt", "value": (saltPPM * 50), "unit": "ppm"}

    pHTank, offset = getSome("i", buff, offset)
    sensors["ph_supply_level"] = {"name": "pH Supply Level", "value": pHTank}

    orpTank, offset = getSome("i", buff, offset)
    sensors["orp_supply_level"] = {"name": "ORP Supply Level", "value": orpTank}

    alarm, offset = getSome("i", buff, offset)
    sensors["chem_alarm"] = {
        "name": "Chemistry Alarm",
        "value": alarm,
        "device_type": DEVICE_TYPE.ALARM,
    }
    # print(json.dumps(data, indent=4))



import struct
from .utility import sendReceiveMessage, getSome, getArray
from ..const import code


def request_equipment_config(gateway_socket, data):
    response = sendReceiveMessage(
        gateway_socket, code.EQUIPMENT_QUERY, struct.pack("<2I", 0, 0)
    )
    decode_equipment_config(response, data)


# pylint: disable=unused-variable
def decode_equipment_config(buff, data):
    if "equipment" not in data:
        data["equipment"] = {}

    equip = data["equipment"]

    equip["controllerType"], offset = getSome("B", buff, 0)
    equip["hardwareType"], offset = getSome("B", buff, offset)

    skip1, offset = getSome("B", buff, offset)
    skip2, offset = getSome("B", buff, offset)
    equip["controllerData"], offset = getSome("I", buff, offset)

    equip["versionDataArray"], offset = getArray(buff, offset)
    equip["speedDataArray"], offset = getArray(buff, offset)
    equip["valveDataArray"], offset = getArray(buff, offset)
    equip["remoteDataArray"], offset = getArray(buff, offset)
    equip["sensorDataArray"], offset = getArray(buff, offset)
    equip["delayDataArray"], offset = getArray(buff, offset)
    equip["macrosDataArray"], offset = getArray(buff, offset)
    equip["miscDataArray"], offset = getArray(buff, offset)
    equip["lightDataArray"], offset = getArray(buff, offset)
    equip["flowsDataArray"], offset = getArray(buff, offset)
    equip["sgsDataArray"], offset = getArray(buff, offset)
    equip["spaFlowsDataArray"], offset = getArray(buff, offset)



import socket
import struct
from ..const import ScreenLogicError, header, code


def sendReceiveMessage(gateway_socket, code, message=b""):
    badCodes = []
    try:
        gateway_socket.sendall(makeMessage(code, message))
        while len(badCodes) < 2:
            data = gateway_socket.recv(1024)
            if not data:
                raise ScreenLogicError(
                    "No data received from the gateway. Request code: {}".format(code)
                )
            rcvCode, buff = takeMessage(data)
            if rcvCode == (code + 1):
                return buff
            else:
                badCodes.append(rcvCode)
        raise ScreenLogicError(
            "Failed to receive the expected response from the gateway within 2 tries. Expected: {}. Received: {}.".format(
                code + 1, badCodes
            )
        )
    except socket.timeout:
        raise ScreenLogicError(
            "Failed to receive the expected response from the gateway within timeout. Expected: {}. Received: {}.".format(
                code + 1, badCodes
            )
        )


def encodeMessageString(string):
    data = string.encode()
    length = len(data)
    pad = 4 - (length % 4)  # pad 'x' to multiple of 4
    fmt = "<I" + str(length) + "s" + str(pad) + "x"
    return struct.pack(fmt, length, data)


def decodeMessageString(data):
    # length = len(data)
    size = struct.unpack_from("<I", data, 0)[0]
    return struct.unpack_from("<" + str(size) + "s", data, struct.calcsize("<I"))[
        0
    ].decode("utf-8")


# Protocol header for every (non-datagram) message sent to/from
# the gateway.
# This header describes the first 8 bytes:
# 0          2          4           8                              N
# | MSG CD 1 | MSG CD 2 | Data Size | Message Data (parameters) -> |

# appends a header to an optional already formatted message.
def makeMessage(msgCode2, messageData=b""):
    return struct.pack(
        header.fmt + str(len(messageData)) + "s",
        code.MSG_CODE_1,
        msgCode2,
        len(messageData),
        messageData,
    )


# takes the header off of the pool message and returns just the message part
def takeMessage(message):
    messageBytes = len(message) - header.length
    # pylint: disable=unused-variable
    rcvCode1, rcvCode2, rcvLen, data = struct.unpack(
        header.fmt + str(messageBytes) + "s", message
    )
    if rcvLen != messageBytes:
        raise ScreenLogicError("Data received does not match expected length")
    if rcvCode2 == code.UNKNOWN_ANSWER:
        raise ScreenLogicError(
            "Unexpected response received from the gateway. Received code_unknown."
        )
    return rcvCode2, data  # return raw data


def getSome(want, buff, offset):
    fmt = want if want.startswith(">") else "<" + want
    newoffset = offset + struct.calcsize(fmt)
    return struct.unpack_from(fmt, buff, offset)[0], newoffset


def getString(buff, offset):
    fmtLen = "<I"
    offsetLen = offset + struct.calcsize(fmtLen)
    sLen = struct.unpack_from(fmtLen, buff, offset)[0]
    if sLen % 4 != 0:
        sLen += 4 - sLen % 4

    fmt = "<{}{}".format(sLen, "s")
    newoffset = offsetLen + struct.calcsize(fmt)
    return struct.unpack_from(fmt, buff, offsetLen)[0], newoffset


def getArray(buff, offset):
    itemCount, offset = getSome("I", buff, offset)
    items = [0 for x in range(itemCount)]
    for i in range(itemCount):
        items[i], offset = getSome("B", buff, offset)
    offsetPad = 0
    if itemCount % 4 != 0:
        offsetPad = (4 - itemCount % 4) % 4
        offset += offsetPad
    return items, offset



# flake8: noqa F401
from .button import request_pool_button_press
from .config import request_pool_config
from .equipment import request_equipment_config
from .gateway import request_gateway_version
from .heat import request_set_heat_mode, request_set_heat_setpoint
from .lights import request_pool_lights_command
from .login import connect_to_gateway
from .pump import request_pump_status
from .status import request_pool_status
from .chemistry import request_chemistry
from .scg import request_scg_config



from .const import DATA
from .requests import (
    connect_to_gateway,
    request_gateway_version,
    request_pool_button_press,
    request_pool_config,
    request_pool_lights_command,
    request_pool_status,
    request_pump_status,
    request_set_heat_mode,
    request_set_heat_setpoint,
    request_chemistry,
    request_scg_config,
)


class ScreenLogicGateway:
    def __init__(
        self, ip, port=80, gtype=0, gsubtype=0, name="Unnamed-Screenlogic-Gateway"
    ):
        self.__ip = ip
        self.__port = port
        self.__type = gtype
        self.__subtype = gsubtype
        self.__name = name
        self.__mac = ""
        self.__version = ""
        self.__connected = False
        self.__data = {}

        if self.__ip and self.__port:
            if self._connect():
                self._get_config()
                self._get_status()
                self._get_pumps()
                self._get_chemistry()
                self._get_scg()
                self._disconnect()
        else:
            raise ValueError("Invalid ip or port")

    @property
    def ip(self):
        return self.__ip

    @property
    def port(self):
        return self.__port

    @property
    def name(self):
        return self.__name

    @property
    def mac(self):
        return self.__mac

    @property
    def version(self):
        return self.__version

    def update(self):
        if (self.is_connected or self._connect()) and self.__data:
            self._get_status()
            self._get_pumps()
            self._get_chemistry()
            self._get_scg()
            self._disconnect()

    def get_data(self):
        return self.__data

    def set_circuit(self, circuitID, circuitState):
        if self._is_valid_circuit(circuitID) and self._is_valid_circuit_state(
            circuitState
        ):
            if self.__connected or self._connect():
                if request_pool_button_press(self.__socket, circuitID, circuitState):
                    self._disconnect()
                    return True
        return False

    def set_heat_temp(self, body, temp):
        if self._is_valid_body(body) and self._is_valid_heattemp(body, temp):
            if self.__connected or self._connect():
                if request_set_heat_setpoint(self.__socket, body, temp):
                    self._disconnect()
                    return True
        return False

    def set_heat_mode(self, body, mode):
        if self._is_valid_body(body) and self._is_valid_heatmode(mode):
            if self.__connected or self._connect():
                if request_set_heat_mode(self.__socket, body, mode):
                    self._disconnect()
                    return True
        return False

    def set_color_lights(self, light_command):
        if self.__connected or self._connect():
            if request_pool_lights_command(self.__socket, light_command):
                self._disconnect()
                return True
        return False

    def is_connected(self):
        return self.__connected

    def _connect(self):
        soc_mac = connect_to_gateway(self.__ip, self.__port)
        if soc_mac:
            self.__socket, self.__mac = soc_mac
            self.__version = request_gateway_version(self.__socket)
            if self.__version:
                self.__connected = True
                return True
        return False

    def _disconnect(self):
        self.__connected = False
        if self.__socket:
            self.__socket.close()

    def _get_config(self):
        if self.__connected or self._connect():
            request_pool_config(self.__socket, self.__data)

    def _get_status(self):
        if self.__connected or self._connect():
            request_pool_status(self.__socket, self.__data)

    def _get_pumps(self):
        if self.__connected or self._connect():
            for pID in self.__data[DATA.KEY_PUMPS]:
                if self.__data[DATA.KEY_PUMPS][pID]["data"] != 0:
                    request_pump_status(self.__socket, self.__data, pID)

    def _get_chemistry(self):
        if self.__connected or self._connect():
            request_chemistry(self.__socket, self.__data)

    def _get_scg(self):
        if self.__connected or self._connect():
            request_scg_config(self.__socket, self.__data)

    def _is_valid_circuit(self, circuit):
        return circuit in self.__data[DATA.KEY_CIRCUITS]

    def _is_valid_circuit_state(self, state):
        return state == 0 or state == 1

    def _is_valid_body(self, body):
        return body in self.__data[DATA.KEY_BODIES]

    def _is_valid_heatmode(self, heatmode):
        return 0 <= heatmode < 5

    def _is_valid_heattemp(self, body, temp):
        min_temp = self.__data[DATA.KEY_BODIES][int(body)]["min_set_point"]["value"]
        max_temp = self.__data[DATA.KEY_BODIES][int(body)]["max_set_point"]["value"]
        return min_temp <= temp <= max_temp



import struct
from .utility import sendReceiveMessage
from ..const import code


def request_set_heat_setpoint(gateway_socket, body, setpoint):
    response = sendReceiveMessage(
        gateway_socket, code.SETHEATTEMP_QUERY, struct.pack("<III", 0, body, setpoint)
    )
    if response == b"":
        return True
    else:
        return False


def request_set_heat_mode(gateway_socket, body, heat_mode):
    response = sendReceiveMessage(
        gateway_socket, code.SETHEATMODE_QUERY, struct.pack("<III", 0, body, heat_mode)
    )
    if response == b"":
        return True
    else:
        return False




# import json
import struct
from .utility import sendReceiveMessage, getSome
from ..const import code, DATA, DEVICE_TYPE


def request_pump_status(gateway_socket, data, pumpID):
    response = sendReceiveMessage(
        gateway_socket, code.PUMPSTATUS_QUERY, struct.pack("<II", 0, pumpID)
    )
    decode_pump_status(response, data, pumpID)


# pylint: disable=unused-variable
def decode_pump_status(buff, data, pumpID):
    # print(buff)
    if DATA.KEY_PUMPS not in data:
        data[DATA.KEY_PUMPS] = {}

    if pumpID not in data[DATA.KEY_PUMPS]:
        data[DATA.KEY_PUMPS][pumpID] = {}

    pump = data[DATA.KEY_PUMPS][pumpID]
    pump["name"] = ""
    pump["pumpType"], offset = getSome("I", buff, 0)
    pump["state"], offset = getSome("I", buff, offset)

    curW, offset = getSome("I", buff, offset)
    pump["currentWatts"] = {}
    curR, offset = getSome("I", buff, offset)
    pump["currentRPM"] = {}

    unknown1, offset = getSome("I", buff, offset)

    curG, offset = getSome("I", buff, offset)
    pump["currentGPM"] = {}

    unknown2, offset = getSome("I", buff, offset)

    pump["presets"] = {}
    name = "Default"
    for i in range(8):
        pump["presets"][i] = {}
        pump["presets"][i]["cid"], offset = getSome("I", buff, offset)
        if DATA.KEY_CIRCUITS in data:
            for num, circuit in data[DATA.KEY_CIRCUITS].items():
                if (
                    pump["presets"][i]["cid"] == circuit["device_id"]
                    and name == "Default"
                ):
                    name = circuit["name"]
                    break
        pump["presets"][i]["setPoint"], offset = getSome("I", buff, offset)
        pump["presets"][i]["isRPM"], offset = getSome("I", buff, offset)

    name = name.strip().strip(",") + " Pump"
    pump["name"] = name

    pump["currentWatts"] = {
        "name": pump["name"] + " Current Watts",
        "value": curW,
        "unit": "W",
        "device_type": DEVICE_TYPE.ENERGY,
    }

    pump["currentRPM"] = {
        "name": pump["name"] + " Current RPM",
        "value": curR,
        "unit": "rpm",
    }

    pump["currentGPM"] = {
        "name": pump["name"] + " Current GPM",
        "value": curG,
        "unit": "gpm",
    }
    # print(json.dumps(data, indent=4))



import struct

# import json
from .utility import sendReceiveMessage, getSome, getString
from ..const import code, BODY_TYPE, DATA


def request_pool_config(gateway_socket, data):
    response = sendReceiveMessage(
        gateway_socket, code.CTRLCONFIG_QUERY, struct.pack("<2I", 0, 0)
    )
    decode_pool_config(response, data)


def decode_pool_config(buff, data):
    # print(buff)
    if DATA.KEY_CONFIG not in data:
        data[DATA.KEY_CONFIG] = {}

    config = data[DATA.KEY_CONFIG]

    controllerID, offset = getSome("I", buff, 0)
    config["controller_id"] = {"name": "Controller ID", "value": controllerID}

    if DATA.KEY_BODIES not in data:
        data[DATA.KEY_BODIES] = {}

    bodies = data[DATA.KEY_BODIES]

    for i in range(2):
        if i not in bodies:
            bodies[i] = {}

        currentBody = bodies[i]

        minSetPoint, offset = getSome("B", buff, offset)
        MINspName = "{} Minimum Set Point".format(BODY_TYPE.NAME_FOR_NUM[i])
        currentBody["min_set_point"] = {"name": MINspName, "value": minSetPoint}
        maxSetPoint, offset = getSome("B", buff, offset)
        MAXspName = "{} Minimum Set Point".format(BODY_TYPE.NAME_FOR_NUM[i])
        currentBody["max_set_point"] = {"name": MAXspName, "value": maxSetPoint}

    degC, offset = getSome("B", buff, offset)
    config["is_celsius"] = {"name": "Is Celsius", "value": degC}

    controllerType, offset = getSome("B", buff, offset)
    config["controller_type"] = controllerType

    hwType, offset = getSome("B", buff, offset)
    config["hardware_type"] = hwType

    controllerbuff, offset = getSome("B", buff, offset)
    config["controller_buffer"] = controllerbuff

    equipFlags, offset = getSome("I", buff, offset)
    config["equipment_flags"] = equipFlags

    paddedGenName, offset = getString(buff, offset)
    genCircuitName = paddedGenName.decode("utf-8").strip("\0")
    config["generic_circuit_name"] = {
        "name": "Default Circuit Name",
        "value": genCircuitName,
    }

    circuitCount, offset = getSome("I", buff, offset)
    config["circuit_count"] = {"name": "Number of Circuits", "value": circuitCount}

    if DATA.KEY_CIRCUITS not in data:
        data[DATA.KEY_CIRCUITS] = {}

    circuits = data[DATA.KEY_CIRCUITS]

    for i in range(circuitCount):

        circuitID, offset = getSome("i", buff, offset)

        if circuitID not in data[DATA.KEY_CIRCUITS]:
            circuits[circuitID] = {}

        currentCircuit = circuits[circuitID]

        currentCircuit["id"] = circuitID

        paddedName, offset = getString(buff, offset)
        circuitName = paddedName.decode("utf-8").strip("\0")
        currentCircuit["name"] = circuitName

        cNameIndex, offset = getSome("B", buff, offset)
        currentCircuit["name_index"] = cNameIndex

        cFunction, offset = getSome("B", buff, offset)
        currentCircuit["function"] = cFunction

        cInterface, offset = getSome("B", buff, offset)
        currentCircuit["interface"] = cInterface

        cFlags, offset = getSome("B", buff, offset)
        currentCircuit["flags"] = cFlags

        cColorSet, offset = getSome("B", buff, offset)
        currentCircuit["color_set"] = cColorSet

        cColorPos, offset = getSome("B", buff, offset)
        currentCircuit["color_position"] = cColorPos

        cColorStagger, offset = getSome("B", buff, offset)
        currentCircuit["color_stagger"] = cColorStagger

        cDeviceID, offset = getSome("B", buff, offset)
        currentCircuit["device_id"] = cDeviceID

        cDefaultRT, offset = getSome("H", buff, offset)
        currentCircuit["default_rt"] = cDefaultRT

        offset = offset + struct.calcsize("2B")

    colorCount, offset = getSome("I", buff, offset)
    config["color_count"] = {"name": "Number of Colors", "value": colorCount}

    if (
        DATA.KEY_COLORS not in data[DATA.KEY_CONFIG]
        or len(config[DATA.KEY_COLORS]) != colorCount
    ):
        config[DATA.KEY_COLORS] = [{} for x in range(colorCount)]

    for i in range(colorCount):
        paddedColorName, offset = getString(buff, offset)
        colorName = paddedColorName.decode("utf-8").strip("\0")
        rgbR, offset = getSome("I", buff, offset)
        rgbG, offset = getSome("I", buff, offset)
        rgbB, offset = getSome("I", buff, offset)
        config[DATA.KEY_COLORS][i] = {"name": colorName, "value": (rgbR, rgbG, rgbB)}

    pumpCircuitCount = 8
    if DATA.KEY_PUMPS not in data:
        data[DATA.KEY_PUMPS] = {}

    pumps = data[DATA.KEY_PUMPS]

    for i in range(pumpCircuitCount):
        if i not in pumps:
            pumps[i] = {}
        pumpData, offset = getSome("B", buff, offset)
        pumps[i]["data"] = pumpData

    interfaceTabFlags, offset = getSome("I", buff, offset)
    config["interface_tab_flags"] = interfaceTabFlags

    showAlarms, offset = getSome("I", buff, offset)
    config["show_alarms"] = showAlarms
    # print(json.dumps(data, indent=4))



import struct
from .utility import sendReceiveMessage
from ..const import code


def request_pool_lights_command(gateway_socket, light_command):
    response = sendReceiveMessage(
        gateway_socket, code.LIGHTCOMMAND_QUERY, struct.pack("<II", 0, light_command)
    )
    if response == b"":
        return True
    else:
        return False



# import json
import struct
from ..const import CHEMISTRY, code, DATA, DEVICE_TYPE, ON_OFF, UNIT
from .utility import sendReceiveMessage, getSome

ADD_UNKNOWNS = False


def request_chemistry(gateway_socket, data):
    response = sendReceiveMessage(
        gateway_socket, code.CHEMISTRY_QUERY, struct.pack("<I", 0)
    )
    decode_chemistry(response, data)


def is_set(bits, mask) -> bool:
    return True if (bits & mask) == mask else False


# pylint: disable=unused-variable
def decode_chemistry(buff, data):
    # print(buff)

    if DATA.KEY_CHEMISTRY not in data:
        data[DATA.KEY_CHEMISTRY] = {}

    chemistry = data[DATA.KEY_CHEMISTRY]

    unit_txt = (
        UNIT.CELSIUS
        if DATA.KEY_CONFIG in data
        and "is_celsius" in data[DATA.KEY_CONFIG]
        and data[DATA.KEY_CONFIG]["is_celsius"]["value"]
        else UNIT.FAHRENHEIT
    )

    unknown = {}

    size, offset = getSome("I", buff, 0)
    unknown["size"] = size

    # skip an unknown value
    unknown1, offset = getSome("B", buff, offset)  # 0
    unknown["unknown1"] = unknown1

    pH, offset = getSome(">H", buff, offset)  # 1
    chemistry["current_ph"] = {"name": "Current pH", "value": (pH / 100), "unit": "pH"}

    orp, offset = getSome(">H", buff, offset)  # 3
    chemistry["current_orp"] = {"name": "Current ORP", "value": orp, "unit": "mV"}

    pHSetpoint, offset = getSome(">H", buff, offset)  # 5
    chemistry["ph_setpoint"] = {
        "name": "pH Setpoint",
        "value": (pHSetpoint / 100),
        "unit": "pH",
    }

    orpSetpoint, offset = getSome(">H", buff, offset)  # 7
    chemistry["orp_setpoint"] = {
        "name": "ORP Setpoint",
        "value": orpSetpoint,
        "unit": "mV",
    }

    pHDoseTime, offset = getSome(">I", buff, offset)  # 9
    chemistry["ph_last_dose_time"] = {
        "name": "Last pH Dose Time",
        "value": pHDoseTime,
        "unit": "Sec" 
    }
    
    orpDoseTime, offset = getSome(">I", buff, offset)  # 13
    chemistry["orp_last_dose_time"] = {
        "name": "Last ORP Dose Time",
        "value": orpDoseTime,
        "unit": "Sec" 
    }

    pHDoseVolume, offset = getSome(">H", buff, offset)  # 17
    chemistry["ph_last_dose_volume"] = {
        "name": "Last pH Dose Volume",
        "value": pHDoseVolume,
        "unit": "mL" 
    }
    
    orpDoseVolume, offset = getSome(">H", buff, offset)  # 19
    chemistry["orp_last_dose_volume"] = {
        "name": "Last ORP Dose Volume",
        "value": orpDoseVolume,
        "unit": "mL" 
    }

    pHSupplyLevel, offset = getSome("B", buff, offset)  # 21 (20)
    chemistry["ph_supply_level"] = {"name": "pH Supply Level", "value": pHSupplyLevel}

    orpSupplyLevel, offset = getSome("B", buff, offset)  # 22 (21)
    chemistry["orp_supply_level"] = {
        "name": "ORP Supply Level",
        "value": orpSupplyLevel,
    }

    saturation, offset = getSome("B", buff, offset)  # 23
    chemistry["saturation"] = {
        "name": "Saturation Index",
        "value": (saturation - 256) / 100
        if is_set(saturation, 0x80)
        else saturation / 100,
        "unit": "lsi",
    }

    cal, offset = getSome(">H", buff, offset)  # 24
    chemistry["calcium_harness"] = {
        "name": "Calcium Hardness",
        "value": cal,
        "unit": "ppm",
    }

    cya, offset = getSome(">H", buff, offset)  # 26
    chemistry["cya"] = {"name": "Cyanuric Acid", "value": cya, "unit": "ppm"}

    alk, offset = getSome(">H", buff, offset)  # 28
    chemistry["total_alkalinity"] = {
        "name": "Total Alkalinity",
        "value": alk,
        "unit": "ppm",
    }

    saltPPM, offset = getSome("B", buff, offset)  # 30
    chemistry["salt_ppm"] = {"name": "Salt", "value": (saltPPM * 50), "unit": "ppm"}

    # Probe temp unit is Celsius?
    probIsC, offset = getSome("B", buff, offset)
    unknown["probe_is_celsius"] = probIsC

    waterTemp, offset = getSome("B", buff, offset)  # 32
    chemistry["ph_probe_water_temp"] = {
        "name": "pH Probe Water Temperature",
        "value": waterTemp,
        "unit": unit_txt,
        "device_type": DEVICE_TYPE.TEMPERATURE,
    }

    if DATA.KEY_ALERTS not in chemistry:
        chemistry[DATA.KEY_ALERTS] = {}

    alerts = chemistry[DATA.KEY_ALERTS]

    alarms, offset = getSome("B", buff, offset)  # 33 (32)
    alerts["flow_alarm"] = {
        "name": "Flow Alarm",
        "value": ON_OFF.from_bool(is_set(alarms, CHEMISTRY.FLAG_ALARM_FLOW)),
    }
    alerts["ph_alarm"] = {
        "name": "pH Alarm",
        "value": ON_OFF.from_bool(is_set(alarms, CHEMISTRY.FLAG_ALARM_PH)),
    }
    alerts["orp_alarm"] = {
        "name": "ORP Alarm",
        "value": ON_OFF.from_bool(is_set(alarms, CHEMISTRY.FLAG_ALARM_ORP)),
    }
    alerts["ph_supply_alarm"] = {
        "name": "pH Supply Alarm",
        "value": ON_OFF.from_bool(is_set(alarms, CHEMISTRY.FLAG_ALARM_PH_SUPPLY)),
    }
    alerts["orp_supply_alarm"] = {
        "name": "ORP Supply Alarm",
        "value": ON_OFF.from_bool(is_set(alarms, CHEMISTRY.FLAG_ALARM_ORP_SUPPLY)),
    }
    alerts["probe_fault_alarm"] = {
        "name": "Probe Fault",
        "value": ON_OFF.from_bool(is_set(alarms, CHEMISTRY.FLAG_ALARM_PROBE_FAULT)),
    }

    if DATA.KEY_NOTIFICATIONS not in chemistry:
        chemistry[DATA.KEY_NOTIFICATIONS] = {}

    notifications = chemistry[DATA.KEY_NOTIFICATIONS]

    warnings, offset = getSome("B", buff, offset)  # 34
    unknown["warnings"] = warnings
    notifications["ph_lockout"] = {
        "name": "pH Lockout",
        "value": ON_OFF.from_bool(is_set(warnings, CHEMISTRY.FLAG_WARNING_PH_LOCKOUT)),
    }
    notifications["ph_limit"] = {
        "name": "pH Daily Limit Reached",
        "value": ON_OFF.from_bool(is_set(warnings, CHEMISTRY.FLAG_WARNING_PH_LIMIT)),
    }
    notifications["orp_limit"] = {
        "name": "ORP Daily Limit Reached",
        "value": ON_OFF.from_bool(is_set(warnings, CHEMISTRY.FLAG_WARNING_ORP_LIMIT)),
    }

    status, offset = getSome("B", buff, offset)  # 35 (34)
    chemistry["status"] = status
    # notifications["corrosive"] = {
    #    "name": "Corrosive",
    #    "value": ON_OFF.from_bool(is_set(status, CHEMISTRY.FLAG_STATUS_CORROSIVE)),
    # }
    # notifications["scaling"] = {
    #    "name": "Scaling",
    #    "value": ON_OFF.from_bool(is_set(status, CHEMISTRY.FLAG_STATUS_SCALING)),
    # }

    chemistry["ph_dosing_state"] = {
        "name": "pH Dosing State",
        "value": (status & CHEMISTRY.MASK_STATUS_PH_DOSING) >> 4,
    }
    chemistry["orp_dosing_state"] = {
        "name": "ORP Dosing State",
        "value": (status & CHEMISTRY.MASK_STATUS_ORP_DOSING) >> 6,
    }

    flags, offset = getSome("B", buff, offset)  # 36 (35)
    chemistry["flags"] = flags

    vMinor, offset = getSome("B", buff, offset)  # 37 (36)
    vMajor, offset = getSome("B", buff, offset)  # 38 (37)
    chemistry["firmware"] = {
        "name": "IntelliChem Firmware Version",
        "value": f"{vMajor}.{str(vMinor).zfill(3)}",
    }

    chemWarnings, offset = getSome("B", buff, offset)  # 39 (38)
    unknown["warnings"] = chemWarnings

    last2, offset = getSome("B", buff, offset)  # 40
    unknown["last2"] = last2
    last3, offset = getSome("B", buff, offset)  # 41
    unknown["last3"] = last3
    last4, offset = getSome("B", buff, offset)  # 42
    unknown["last4"] = last4

    if ADD_UNKNOWNS:
        chemistry["unknown"] = unknown

    # print(json.dumps(data, indent=4))
